package com.loganmagnan.pluginbase;

import com.loganmagnan.pluginbase.menusystem.PlayerMenuUtil;
import com.loganmagnan.pluginbase.utils.ClassRegistrationUtils;
import com.loganmagnan.pluginbase.utils.ColorUtils;
import com.loganmagnan.pluginbase.utils.Utils;
import com.loganmagnan.pluginbase.utils.command.CommandFramework;
import com.loganmagnan.pluginbase.utils.config.FileConfig;
import com.loganmagnan.pluginbase.utils.config.file.Config;
import lombok.Getter;
import lombok.Setter;
import org.bukkit.entity.Player;
import org.bukkit.plugin.java.JavaPlugin;
import java.util.HashMap;

@Getter
@Setter
public class PluginBase extends JavaPlugin {

    // Main class instance
    @Getter private static PluginBase instance;

    // Configuration files
    private Config mainConfig;
    private FileConfig messagesConfig;

    // Managers

    // Menu system
    private HashMap<Player, PlayerMenuUtil> playerMenuUtilMap = new HashMap<>();

    // Command framework
    private CommandFramework commandFramework = new CommandFramework(this);

    // On enable function
    @Override
    public void onEnable() {
        // Initialize the main class instance
        instance = this;

        // Initialize all of the configuration files
        this.saveDefaultConfig();
        this.mainConfig = new Config("config", this, true);
        this.messagesConfig = new FileConfig(this, "messages.yml", true);

        // Say the plugin's name
        this.getServer().getConsoleSender().sendMessage(Utils.chatBar);
        this.getServer().getConsoleSender().sendMessage(ColorUtils.getMessageType("&dPluginBase &7- &av" + this.getDescription().getVersion()));
        this.getServer().getConsoleSender().sendMessage(ColorUtils.getMessageType("&7Made by &eLoganM Development"));
        this.getServer().getConsoleSender().sendMessage(Utils.chatBar);

        // Load all of the functions
        this.loadCommands();
        this.loadManagers();
        this.loadListeners();
        this.loadRunnables();
    }

    // On disable function
    @Override
    public void onDisable() {
        // Remove the main class instance
        instance = null;
    }

    // Load commands function
    private void loadCommands() {
        ClassRegistrationUtils.loadCommands("com.loganmagnan.pluginbase.commands");
    }

    // Load managers function
    private void loadManagers() {

    }

    // Load listeners function
    private void loadListeners() {
        ClassRegistrationUtils.loadListeners("com.loganmagnan.pluginbase.listeners");
    }

    // Load runnables function
    private void loadRunnables() {

    }

    // Menu system function
    public PlayerMenuUtil getPlayerMenuUtil(Player player) {
        PlayerMenuUtil playerMenuUtil;

        if (playerMenuUtilMap.containsKey(player)) {
            return playerMenuUtilMap.get(player);
        } else {
            playerMenuUtil = new PlayerMenuUtil(player);

            playerMenuUtilMap.put(player, playerMenuUtil);

            return playerMenuUtil;
        }
    }
}
